#!/usr/bin/env python

import json
import logging
import os
import pathlib
import platform
import shutil
from dataclasses import dataclass, field
from logging.handlers import TimedRotatingFileHandler
from typing import List

from dacite import from_dict

from nidibot.bots.bot_factory import BotFactory
from nidibot.bots.bot_interface import BotConfiguration, BotInterface
from nidibot.server_provider.game_server import GameServer
from nidibot.server_provider.server_provider_factory import ServerProviderFactory
from nidibot.server_provider.server_provider_interface import (
    ServerProviderConfiguration,
    ServerProviderInterface,
)


@dataclass
class GeneralConfiguration:
    backups_folder_path: str = ""
    logs_folder_path: str = ""


@dataclass
class NidibotConfiguration:
    general: GeneralConfiguration = field(default_factory=GeneralConfiguration)
    bots: List[BotConfiguration] = field(default_factory=list)
    server_providers: List[ServerProviderConfiguration] = field(default_factory=list)


class Nidibot:
    def __init__(self, working_folder_path: str):
        if not working_folder_path:
            raise ValueError("Working directory value is required for nidibot start!")

        self.__working_folder_path = working_folder_path
        self.__configuration: NidibotConfiguration = (
            self.__parse_configuration_json_file()
        )

        self.__initialize_logging()

        root_backup_path = os.path.join(self.__working_folder_path, "backups")
        if os.path.isabs(self.__configuration.general.backups_folder_path):
            root_backup_path = self.__configuration.general.backups_folder_path

        pathlib.Path(root_backup_path).mkdir(parents=True, exist_ok=True)

        self.__bots: List[BotInterface] = []

        self.__server_providers: List[ServerProviderInterface] = (
            ServerProviderFactory.create_all(
                configuration_list=self.__configuration.server_providers,
                root_backup_path=root_backup_path,
                notify_callback=self.__notify_callback,
            )
        )
        if len(self.__server_providers) == 0:
            raise ValueError("At least one server provider is required!")

        self.__game_servers: List[GameServer] = []
        for server_provider in self.__server_providers:
            self.__game_servers.extend(server_provider.get_servers())

        self.__bots = BotFactory.create_all(
            configuration_list=self.__configuration.bots,
            game_servers=self.__game_servers,
        )
        if len(self.__bots) == 0:
            raise ValueError("At least one bot is required!")

    def __parse_configuration_json_file(self) -> NidibotConfiguration:
        configuration_filepath = os.path.join(
            self.__working_folder_path, "bot_configuration.json"
        )

        with open(configuration_filepath, "r", encoding="utf-8") as json_file:
            configuration_json = json.load(json_file)

        configuration = from_dict(
            data_class=NidibotConfiguration,
            data=configuration_json,
        )

        return configuration

    def __initialize_logging(self) -> None:
        #
        # Enable daily logging both to file and stdout.
        #
        log_directory = os.path.join(self.__working_folder_path, "logs")
        if os.path.isabs(self.__configuration.general.logs_folder_path):
            log_directory = self.__configuration.general.logs_folder_path

        pathlib.Path(log_directory).mkdir(parents=True, exist_ok=True)

        filename = "nidibot.log"
        filepath = os.path.join(log_directory, filename)

        handler = TimedRotatingFileHandler(filepath, when="midnight", backupCount=60)
        handler.suffix = "%Y%m%d"

        logging.basicConfig(
            level=logging.DEBUG,
            format="%(asctime)s:%(levelname)s:%(funcName)s(): %(message)s",
            handlers=[handler, logging.StreamHandler()],
        )

        logging.getLogger("hikari").setLevel(logging.WARNING)
        logging.getLogger("lightbulb").setLevel(logging.WARNING)
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)

    def __notify_callback(self, title: str, message: str):
        for bot in self.__bots:
            bot.notify(title, message)

    def start(self) -> None:
        for bot in self.__bots:
            bot.activate()

    @staticmethod
    def initialize_folder() -> None:
        """ """
        current_script_folder = os.path.dirname(os.path.realpath(__file__))
        current_working_folder = os.getcwd()
        shutil.copytree(
            os.path.join(current_script_folder, "templates", "common"),
            current_working_folder,
            dirs_exist_ok=True,
        )

        if platform.system() == "Linux":
            shutil.copytree(
                os.path.join(current_script_folder, "templates", "linux"),
                current_working_folder,
                dirs_exist_ok=True,
            )

            #
            # Modify path to working folder in "nidibot.service" file.
            #
            file_lines: list = []
            service_filepath = os.path.join(current_working_folder, "nidibot.service")
            with open(
                file=service_filepath, mode="r", encoding="utf-8"
            ) as service_file:
                file_lines = service_file.readlines()

            for line_index, line in enumerate(file_lines):
                if "ExecStart=/user/bin/python3 /home/nidibot/start_bot.py" in line:
                    script_filepath = os.path.join(
                        current_working_folder, "start_bot.py"
                    )
                    edited_line = f"ExecStart=/user/bin/python3 {script_filepath}\n"
                    file_lines[line_index] = edited_line

            with open(
                file=service_filepath, mode="w", encoding="utf-8"
            ) as service_file:
                service_file.writelines(file_lines)



#!/usr/bin/env python

import logging
from dataclasses import dataclass
from datetime import date, datetime
from multiprocessing import Lock
from typing import List

import hikari
import lightbulb
import pkg_resources  # type: ignore
from lightbulb.ext import tasks

from nidibot.bots.bot_interface import BotConfiguration, BotInterface
from nidibot.server_provider.game_server import GameServer


@dataclass
class NotifyMessage:
    title: str = ""
    message: str = ""


class DiscordBot(BotInterface):
    def __init__(self, configuration: BotConfiguration, game_servers: List[GameServer]):
        self.__configuration = configuration
        self.__game_servers = game_servers

        self.__game_server_names: list = []
        for game_server in self.__game_servers:
            self.__game_server_names.append(game_server.name())

        self.__notify_mutex = Lock()
        self.__notify_messages: List[NotifyMessage] = []

        self.__bot = lightbulb.BotApp(token=self.__configuration.token)
        tasks.load(self.__bot)

        self.__color_green = 0x37CB78
        self.__color_orange = 0xE67E22
        self.__color_red = 0xE64A42

        def get_embed_title(game_server) -> str:
            server_status = game_server.status()

            title = f"{server_status.game_name}"
            if server_status.version:
                title += f" ({server_status.version})"

            title += f" - {server_status.address}"

            return title

        @self.__bot.listen(hikari.StartedEvent)
        async def on_started(_) -> None:
            try:
                nidibot_version = pkg_resources.get_distribution("nidibot").version
                logging.info("nidibot v%s was started.", nidibot_version)
            except pkg_resources.DistributionNotFound:
                logging.debug("nidibot was started.")

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self.__game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="status",
            description="Provides extended information about game server status.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def status(ctx) -> None:
            logging.debug("Called 'status' by '%s'.", ctx.author)

            server_name = ctx.options.name
            if server_name is None:
                game_server = self.__game_servers[0]
            else:
                game_server = next(
                    x for x in self.__game_servers if x.name() == server_name
                )

            server_status = game_server.status()

            title = get_embed_title(game_server)

            if server_status.status == "online":
                status_smiley = ":white_check_mark:"
                status_color = hikari.colors.Color(self.__color_green)
            elif server_status.status == "offline":
                status_smiley = ":no_entry:"
                status_color = hikari.colors.Color(self.__color_red)
            elif server_status.status == "restarting":
                status_smiley = ":warning:"
                status_color = hikari.colors.Color(self.__color_orange)
            else:
                status_smiley = ":interrobang:"
                status_color = hikari.colors.Color(self.__color_orange)

            embed = hikari.Embed(
                title=title,
                color=status_color,
            )

            embed.add_field(
                name="Address:", value=f"`{server_status.address}`", inline=True
            )
            embed.add_field(
                name="Status:",
                value=f"{status_smiley} {server_status.status}",
                inline=True,
            )

            players = (
                f"{server_status.players_connected} / {server_status.players_limit}"
            )
            if len(server_status.player_names) > 0:
                players += f" ({server_status.player_names})"
            embed.add_field(name="Players:", value=f"{players}", inline=True)

            update_smiley = ""
            if server_status.update_available:
                update_smiley = ":warning:"
            else:
                update_smiley = ":white_check_mark:"

            date_time = server_status.available_until.split(" ")
            date_parts = date_time[0].split("-")

            d0 = date(datetime.now().year, datetime.now().month, datetime.now().day)
            d1 = date(int(date_parts[0]), int(date_parts[1]), int(date_parts[2]))
            delta = d1 - d0

            embed.add_field(
                name="Available until:",
                value=f"{server_status.available_until} ({delta.days} days left)",
                inline=True,
            )
            embed.add_field(
                name="Update available:",
                value=f"{update_smiley} {'yes' if server_status.update_available else 'no'}",
                inline=True,
            )

            await ctx.respond(embed=embed)

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self.__game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="start",
            description="Starts server if it is offline, restarts server if it is online.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def start(ctx) -> None:
            logging.debug("Called 'start' by '%s'.", ctx.author)

            server_name = ctx.options.name
            if server_name is None:
                game_server = self.__game_servers[0]
            else:
                game_server = next(
                    x for x in self.__game_servers if x.name() == server_name
                )

            title = get_embed_title(game_server)

            user = str(ctx.author)
            if user not in self.__configuration.privileged_users:
                embed = hikari.Embed(
                    title=title,
                    description="Sorry but you don't have rights to call this command! :liar:",
                    color=hikari.colors.Color(self.__color_red),
                )

                await ctx.respond(embed=embed)
                return

            embed = hikari.Embed(
                title=title,
                description=":warning: Starting server! :warning:",
                color=hikari.colors.Color(self.__color_red),
            )
            await ctx.respond(embed=embed)

            game_server.start()

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self.__game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="stop",
            description="Stops server if it is online.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def stop(ctx) -> None:
            logging.debug("Called 'stop' by '%s'.", ctx.author)

            server_name = ctx.options.name
            if server_name is None:
                game_server = self.__game_servers[0]
            else:
                game_server = next(
                    x for x in self.__game_servers if x.name() == server_name
                )

            title = get_embed_title(game_server)

            user = str(ctx.author)
            if user not in self.__configuration.privileged_users:
                embed = hikari.Embed(
                    title=title,
                    description="Sorry but you don't have rights to call this command! :liar:",
                    color=hikari.colors.Color(self.__color_red),
                )

                await ctx.respond(embed=embed)
                return

            embed = hikari.Embed(
                title=title,
                description=":warning: Stopping server! :warning:",
                color=hikari.colors.Color(self.__color_red),
            )
            await ctx.respond(embed=embed)

            game_server.stop()

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self.__game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="restart",
            description="Restarts server if it is online, starts server if it is offline.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def restart(ctx) -> None:
            logging.debug("Called 'restart' by '%s'.", ctx.author)

            server_name = ctx.options.name
            if server_name is None:
                game_server = self.__game_servers[0]
            else:
                game_server = next(
                    x for x in self.__game_servers if x.name() == server_name
                )

            title = get_embed_title(game_server)

            user = str(ctx.author)
            if user not in self.__configuration.privileged_users:
                embed = hikari.Embed(
                    title=title,
                    description="Sorry but you don't have rights to call this command! :liar:",
                    color=hikari.colors.Color(self.__color_red),
                )

                await ctx.respond(embed=embed)
                return

            embed = hikari.Embed(
                title=title,
                description=":warning: Restarting server! :warning:",
                color=hikari.colors.Color(self.__color_red),
            )
            await ctx.respond(embed=embed)

            game_server.restart()

        @self.__bot.command
        @lightbulb.option(
            name="name",
            description="States server to which command will be applied",
            choices=self.__game_server_names,
            required=False,
        )
        @lightbulb.command(
            name="backup",
            description="Creates backup of games server files and uploads them to storage.",
        )
        @lightbulb.implements(lightbulb.SlashCommand)
        async def backup(ctx) -> None:
            logging.debug("Called 'backup' by '%s'.", ctx.author)

            server_name = ctx.options.name
            if server_name is None:
                game_server = self.__game_servers[0]
            else:
                game_server = next(
                    x for x in self.__game_servers if x.name() == server_name
                )

            title = get_embed_title(game_server)

            user = str(ctx.author)
            if user not in self.__configuration.privileged_users:
                embed = hikari.Embed(
                    title=title,
                    description="Sorry but you don't have rights to call this command! :liar:",
                    color=hikari.colors.Color(self.__color_red),
                )

                await ctx.respond(embed=embed)
                return

            embed = hikari.Embed(
                title=title,
                description=":warning: Started creating backup of the server, please wait.",
                color=hikari.colors.Color(self.__color_orange),
            )
            await ctx.respond(embed=embed)

            if game_server.create_backup():
                embed = hikari.Embed(
                    title=title,
                    description=":white_check_mark: Backup was created successfully!",
                    color=hikari.colors.Color(self.__color_green),
                )
            else:
                embed = hikari.Embed(
                    title=title,
                    description=":no_entry: Backup creation failed, please check bot logs!",
                    color=hikari.colors.Color(self.__color_red),
                )

            await ctx.respond(embed=embed)

        @tasks.task(s=5, auto_start=True)
        async def notify_loop():
            local_notify_messages: List[NotifyMessage] = []
            with self.__notify_mutex:
                local_notify_messages = self.__notify_messages
                self.__notify_messages = []

            if len(local_notify_messages) == 0:
                return

            connected_channels: list = []
            connected_guilds = await self.__bot.rest.fetch_my_guilds()
            for guild in connected_guilds:
                channels = await self.__bot.rest.fetch_guild_channels(guild)
                for channel in channels:
                    if channel.type == hikari.ChannelType.GUILD_TEXT:
                        connected_channels.append(channel)

            for notify_message in local_notify_messages:
                embed = hikari.Embed(
                    title=notify_message.title,
                    description=f":warning: {notify_message.message}",
                    color=hikari.colors.Color(self.__color_orange),
                )

                for channel in connected_channels:
                    try:
                        await self.__bot.rest.create_message(
                            channel=channel.id, embed=embed
                        )
                    except hikari.errors.ForbiddenError as exception:
                        logging.exception(exception)

    def notify(self, title: str, message: str) -> None:
        with self.__notify_mutex:
            notify_message: NotifyMessage = NotifyMessage()
            notify_message.title = title
            notify_message.message = message
            self.__notify_messages.append(notify_message)

    def activate(self) -> bool:
        try:
            self.__bot.run()
            return True
        except Exception as exception:
            logging.exception(exception)
            return False



#!/usr/bin/env python

from typing import List

from nidibot.bots.bot_interface import BotConfiguration
from nidibot.bots.discord_bot import DiscordBot
from nidibot.server_provider.game_server import GameServer


class BotFactory:
    @staticmethod
    def create(configuration: BotConfiguration, game_servers: List[GameServer]):
        bots = {
            "discord": DiscordBot,
        }

        if configuration.type not in bots:
            if not configuration.type:
                raise ValueError("Empty bot type provided!")

            raise ValueError("Unknown bot type provided!")

        return bots[configuration.type](
            configuration=configuration, game_servers=game_servers
        )

    @staticmethod
    def create_all(configuration_list: list, game_servers: List[GameServer]) -> list:
        bots: list = []

        for configuration in configuration_list:
            bots.append(
                BotFactory.create(
                    configuration=configuration, game_servers=game_servers
                )
            )

        return bots



#!/usr/bin/env python

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List

from nidibot.server_provider.game_server import GameServer


@dataclass
class BotConfiguration:
    type: str = ""
    token: str = ""
    privileged_users: list = field(default_factory=list)


class BotInterface(ABC):
    @abstractmethod
    def __init__(self, configuration: BotConfiguration, game_servers: List[GameServer]):
        pass

    @abstractmethod
    def notify(self, title: str, message: str) -> None:
        pass

    @abstractmethod
    def activate(self) -> bool:
        pass
